#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>
#include <time.h>
#include "irc.h"

#ifdef ENABLE_SSL
	#include <openssl/err.h>
#endif

#define bool char
#define VERSION "2.0.1"

typedef struct {
	char *dir;
	char *ext;
	char *fmt;
	char *host;
	char *nick;
	int port;
	bool usessl;
	bool debug;
} CFG;

char *prog_name;
IRCC_client ircbot;
static CFG cfg = {
	.dir = "./",
	.ext = ".txt",
	.fmt = "[%s] %s",
	.host = "127.0.0.1",
	.nick = "historybot",
	.port = 6667
};

void die(const char *fmt, ...) {
	va_list args;
	va_start(args, fmt);

	vfprintf(stderr, fmt, args);
	va_end(args);

	IRCC_close(&ircbot);
	exit(1);
}

void sig_handler(int sig) {
	fprintf(stderr, "%s: received signal: %d\n", prog_name, sig);
	IRCC_close(&ircbot);
	exit(0);
}

char *GetFilename(void) {
	time_t t = time(NULL);
	struct tm *tm = localtime(&t);
	if (tm == NULL)
		die("%s: localtime(): %s\n", prog_name, strerror(errno));

	static char filename[PATH_MAX + 1];
	snprintf(filename, PATH_MAX, "%s_%d-%d-%d%s",
		ircbot.irc_channel,
		tm->tm_mday,
		tm->tm_mon,
		1900 + tm->tm_year,
		cfg.ext);

	return filename;
}

void WriteToFile(void) {
	/* Secret message */
	if (ircbot.irc_msg[0] == '!')
		return;

	char *filename = GetFilename();
	FILE *fp = fopen(filename, "a");
	if (fp == NULL) {
		fprintf(stderr, "%s: cant open file: %s\n", prog_name, filename);
		return;
	}

	fprintf(fp, cfg.fmt, ircbot.irc_nick, ircbot.irc_msg);
	fclose(fp);
}

int main(int argc, char **argv) {
	prog_name = argv[0];
	bool daemonize = 0;

	int opt;
	while ((opt = getopt(argc, argv, "h:p:d:n:f:e:tDVa")) != -1) {
		switch (opt) {
			case 'h':
				cfg.host = optarg;
				break;

			case 'p':
				cfg.port = atoi(optarg);
				break;

			case 'd':
				cfg.dir = optarg;
				break;

			case 'n':
				cfg.nick = optarg;
				break;

			case 'e':
				cfg.ext = optarg;
				break;

			case 'f':
				cfg.fmt = optarg;
				break;

			case 't':
				cfg.usessl = 1;
				break;

			case 'V':
				printf("%s %s (history) and using irclibs %s\nWritten under WTFPL License\n", prog_name, VERSION, IRCC_VERSION);
				return 0;

			case 'D':
				cfg.debug = 1;
				break;

			case 'a':
				daemonize = 1;
				break;

			default:
				printf("%s -[h:p:d:n:e:f:tDVa] [\"#channel:key\"]\n\t-h STR Server's ip        Default: %s\n\t-p INT Server's port      Default: %d\n\t-d STR Logging dir        Default: %s\n\t-n STR Bot nickname       Default: %s\n\t-e STR File extantion     Default: %s\n\t-f STR Log format         Default: %s\n\t-D Print messages(debug)  Default: %s\n\t-V Print version and license\n\t-a Fork and daemonize\n",
					prog_name,
					cfg.host,
					cfg.port,
					cfg.dir,
					cfg.nick,
					cfg.ext,
					cfg.fmt,
					(cfg.debug) ? "true" : "false");

				#if defined(ENABLE_SSL) || defined(ENABLE_TLS)
					printf("\t-t Use ssl                Default: %s\n", (cfg.usessl) ? "true" : "false");
				#endif

				return 0;
		}
	}

	if (daemonize) {
		pid_t pid = fork();
		if (pid < 0)
			die("%s: fork: %s\n", prog_name, strerror(errno));

		/* Parent */
		else if (pid != 0)
			return 0;

		if (daemon(1, 0) < 0)
			die("%s: daemon: %s\n", prog_name, strerror(errno));
	}

	signal(SIGTERM, sig_handler);
	signal(SIGKILL, sig_handler);
	signal(SIGINT, sig_handler);

	argv += optind;
	argc -= optind;

	if (chdir(cfg.dir))
		die("%s: chdir(): %s\n", prog_name, strerror(errno));

	int status = IRCC_connect(&ircbot, cfg.host, cfg.port);
	if (status < 0)
		die("%s: connect(): %s\n", prog_name, strerror(errno));

	if (cfg.usessl) {
		#if defined(ENABLE_SSL) || defined(ENABLE_TLS)
			if (IRCC_initssl(&ircbot) == IRCC_ERROR)
				die(ERR_error_string(ERR_get_error(), NULL));
		#endif
	}

	/* Register and skip MOTD */
	if (IRCC_register(&ircbot, cfg.nick) < 0)
		die("%s: irc register: %s\n", prog_name, strerror(errno));

	sleep(5);

	/* Join a channel */
	for (int i = 0; i < argc; i++) {
		char *key = "";
		char *channel = argv[i];

		char *p = strchr(channel, ':');
		if (p != NULL) {
			p[0] = '\0';
			key = p + 1;
		}

		IRCC_join(&ircbot, channel, key);
	}

	/* Logging */
	while (1) {
		int irc_status = IRCC_recv(&ircbot);
		if (irc_status < 0)
			die("%s: irc recv: %s\n", prog_name, strerror(errno));

		irc_status = IRCC_parse(&ircbot);
		if (ircbot.irc_nick != NULL && ircbot.irc_channel != NULL && ircbot.irc_msg != NULL && irc_status == IRCC_PRIVMSG)
			WriteToFile();

		if (cfg.debug)
			printf("\033[32m%s\033[0m", ircbot.irc_raw);
	}
}
